
from LogicWeave import LogicWeave
import LogicWeave.proto_gen.logicweave_core_pb2 as logicweave_core_pb2

class LogicWeaveCore(LogicWeave):
    def __init__(self, *args, **kwargs):
        kwargs['protobuf_module'] = logicweave_core_pb2
        super().__init__(*args, **kwargs)

    def read_voltage(self):
        request = self.pb.ReadVoltageRequest()
        return self._send_and_parse(request, "read_voltage_response")

    def read_resistance(self):
        request = self.pb.ReadResistanceRequest()
        return self._send_and_parse(request, "read_resistance_response")
    
    def read_pd(self):
        request = self.pb.ReadPDRequest()
        return self._send_and_parse(request, "read_pd_response")
    
    def set_psu_output(self, channel, state):
        request = self.pb.SetPSUOutputRequest(channel=channel, state=state)
        return self._send_and_parse(request, "set_psu_output_response")
    
    def read_power_monitor(self):
        request = self.pb.ReadPowerMonitorRequest()
        return self._send_and_parse(request, "read_power_monitor_response")
    
    def configure_psu(self, channel, voltage, current_limit):
        request = self.pb.ConfigurePSURequest(channel=channel, voltage=voltage, current_limit=current_limit)
        return self._send_and_parse(request, "configure_psu_response")
    
    def cal_probes(self):
        request = self.pb.ZeroProbesRequest()
        return self._send_and_parse(request, "zero_probes_response")

    def read_calibration_data(self):
        request = self.pb.ReadCalibrationDataRequest()
        return self._send_and_parse(request, "read_calibration_data_response")


import struct
import logging
import sys
import builtins
import enum
from typing import Optional, Any
from LogicWeave.exceptions import DeviceFirmwareError, DeviceResponseError, DeviceConnectionError
import LogicWeave.proto_gen.logicweave_pb2 as lw_pb2
from .transports import Transport, NativeUsbTransport, PyodideTransport, VENDOR_ID, PRODUCT_ID, INTERFACE_NUM, PACKET_SIZE

ProtobufModule = Any 

class GpioFunction(enum.IntEnum):
    """
    Defines the available functions (modes) for a GPIO pin,
    mapping directly to the integer values in the GpioFunction protobuf enum.
    """
    XIP = 0
    HSTX = 1
    SPI = 2
    UART = 3
    I2C = 4
    PWM = 5
    SIO_IN = 6
    SIO_OUT = 7
    PIO = 8
    GPCK = 9
    USB = 10
    NONE = 11

# Default GPIO value for optional, unconfigured pins.
DEFAULT_GPIO_PIN = 63

# --- Base Class for Peripherals ---
class _BasePeripheral:
    """A base class for peripheral controllers to reduce boilerplate."""
    def __init__(self, controller: 'LogicWeave'):
        self._controller = controller
        self.pb: ProtobufModule = controller.pb 

    def _build_and_execute(self, request_class, expected_response_field: str, **kwargs):
        request_payload = request_class(**kwargs)
        return self._controller._send_and_parse(request_payload, expected_response_field)

# --- Peripheral Classes ---
class UART(_BasePeripheral):
    def __init__(self, controller: 'LogicWeave', instance_num: int, 
                 tx_pin: int = DEFAULT_GPIO_PIN, rx_pin: int = DEFAULT_GPIO_PIN, 
                 baud_rate: int = 115200):
        super().__init__(controller)
        self._instance_num = instance_num
        self.tx_pin = tx_pin
        self.rx_pin = rx_pin
        self.baud_rate = baud_rate
        self._setup()

    def _setup(self):
        self._build_and_execute(self.pb.UartSetupRequest, "uart_setup_response", 
                                instance_num=self._instance_num, tx_pin=self.tx_pin, 
                                rx_pin=self.rx_pin, baud_rate=self.baud_rate)

    def write(self, data: bytes, timeout_ms: int = 1000):
        self._build_and_execute(self.pb.UartWriteRequest, "uart_write_response", 
                                instance_num=self._instance_num, data=data, 
                                timeout_ms=timeout_ms)

    def read(self, byte_count: int, timeout_ms: int = 1000) -> bytes:
        response = self._build_and_execute(self.pb.UartReadRequest, "uart_read_response", 
                                           instance_num=self._instance_num, 
                                           byte_count=byte_count, timeout_ms=timeout_ms)
        return response.data

    def __repr__(self):
        return f"<UART instance={self._instance_num} tx={self.tx_pin} rx={self.rx_pin} baud={self.baud_rate}>"

class GPIO(_BasePeripheral):
    MAX_ADC_COUNT = 4095
    V_REF = 3.3

    def __init__(self, controller: 'LogicWeave', pin: int, name: Optional[str] = "gpio"):
        super().__init__(controller)
        self.pin = pin
        self.pull = None
        self.name = name

    def set_function(self, mode: GpioFunction):
        self._build_and_execute(self.pb.GPIOFunctionRequest, "gpio_function_response", 
                                gpio_pin=self.pin, function=mode, name=self.name)

    def set_pull(self, state: int):
        self._build_and_execute(self.pb.GpioPinPullRequest, "gpio_pin_pull_response", 
                                gpio_pin=self.pin, state=state)
        self.pull = state

    def write(self, state: bool):
        if self._controller.read_pin_function(self.pin) != self.pb.GpioFunction.sio_out:
            self.set_function(self.pb.GpioFunction.sio_out)
        self._build_and_execute(self.pb.GPIOWriteRequest, "gpio_write_response", 
                                gpio_pin=self.pin, state=state)

    def read(self) -> bool:
        if self._controller.read_pin_function(self.pin) != self.pb.GpioFunction.sio_in:
            self.set_function(self.pb.GpioFunction.sio_in)
        response = self._build_and_execute(self.pb.GPIOReadRequest, "gpio_read_response", 
                                           gpio_pin=self.pin)
        return response.state

    def setup_pwm(self, wrap, clock_div_int=0, clock_div_frac=0):
        self._build_and_execute(self.pb.PWMSetupRequest, "pwm_setup_response", 
                                gpio_pin=self.pin, wrap=wrap, 
                                clock_div_int=clock_div_int, 
                                clock_div_frac=clock_div_frac, name=self.name)

    def set_pwm_level(self, level):
        self._build_and_execute(self.pb.PWMSetLevelRequest, "pwm_set_level_response", 
                                gpio_pin=self.pin, level=level)

    def read_adc(self) -> float:
        response = self._build_and_execute(self.pb.ADCReadRequest, "adc_read_response", gpio_pin=self.pin)
        return (response.sample / self.MAX_ADC_COUNT) * self.V_REF

    def __repr__(self):
        return f"<GPIO pin={self.pin}>"

class I2C(_BasePeripheral):
    def __init__(self, controller: 'LogicWeave', instance_num: int, 
                 sda_pin: int = DEFAULT_GPIO_PIN, scl_pin: int = DEFAULT_GPIO_PIN, 
                 name: Optional[str] = "i2c"):
        super().__init__(controller)
        self._instance_num = instance_num
        self.sda_pin = sda_pin
        self.scl_pin = scl_pin
        self.name = name
        self._setup()

    def _setup(self):
        self._build_and_execute(self.pb.I2CSetupRequest, "i2c_setup_response", 
                                instance_num=self._instance_num, sda_pin=self.sda_pin, 
                                scl_pin=self.scl_pin, name=self.name)

    def write(self, device_address: int, data: bytes):
        self._build_and_execute(self.pb.I2CWriteRequest, "i2c_write_response", 
                                instance_num=self._instance_num, 
                                device_address=device_address, data=data)

    def write_then_read(self, device_address: int, data: bytes, byte_count: int) -> bytes:
        response = self._build_and_execute(self.pb.I2CWriteThenReadRequest, "i2c_write_then_read_response", 
                                            instance_num=self._instance_num, 
                                            device_address=device_address, data=data, 
                                            byte_count=byte_count)
        return response.data

    def read(self, device_address: int, byte_count: int) -> bytes:
        response = self._build_and_execute(self.pb.I2CReadRequest, "i2c_read_response", 
                                            instance_num=self._instance_num, 
                                            device_address=device_address, 
                                            byte_count=byte_count)
        return response.data

    def __repr__(self):
        return f"<I2C instance={self._instance_num} sda={self.sda_pin} scl={self.scl_pin}>"

class SPI(_BasePeripheral):
    def __init__(self, controller: 'LogicWeave', instance_num: int, 
                 sclk_pin: int, mosi_pin: int, 
                 miso_pin: int = DEFAULT_GPIO_PIN, baud_rate: int = 1e6, 
                 name: Optional[str] = "spi", default_cs_pin: Optional[int] = None):
        super().__init__(controller)
        self._instance_num = instance_num
        self.sclk_pin = sclk_pin
        self.mosi_pin = mosi_pin
        self.miso_pin = miso_pin
        self.baud_rate = baud_rate
        self._default_cs_pin = default_cs_pin
        self.name = name
        self._setup()

    def _setup(self):
        self._build_and_execute(self.pb.SPISetupRequest, "spi_setup_response", 
                                instance_num=self._instance_num, sclk_pin=self.sclk_pin, 
                                mosi_pin=self.mosi_pin, miso_pin=self.miso_pin, 
                                baud_rate=self.baud_rate, name=self.name)

    def _get_cs_pin(self, cs_pin_override: Optional[int]) -> int:
        active_cs_pin = cs_pin_override if cs_pin_override is not None else self._default_cs_pin
        if active_cs_pin is None: 
            raise ValueError("A Chip Select (CS) pin must be provided.")
        return active_cs_pin

    def write(self, data: bytes, cs_pin: Optional[int] = None):
        self._build_and_execute(self.pb.SPIWriteRequest, "spi_write_response", 
                                instance_num=self._instance_num, data=data, 
                                cs_pin=self._get_cs_pin(cs_pin))

    def read(self, byte_count: int, cs_pin: Optional[int] = None, data_to_send: int = 0) -> bytes:
        response = self._build_and_execute(self.pb.SPIReadRequest, "spi_read_response", 
                                            instance_num=self._instance_num, 
                                            data=data_to_send, 
                                            cs_pin=self._get_cs_pin(cs_pin), 
                                            byte_count=byte_count)
        return response.data

    def __repr__(self):
        parts = [f"<SPI instance={self._instance_num}", f"sclk={self.sclk_pin}", f"mosi={self.mosi_pin}", f"miso={self.miso_pin}"]
        if self._default_cs_pin is not None: 
            parts.append(f"default_cs={self._default_cs_pin}")
        return " ".join(parts) + ">"

# --- Main Controller Class ---
class LogicWeave:
    def __init__(self, transport: Transport = None, protobuf_module: ProtobufModule = lw_pb2, 
                 vendor_id: int = VENDOR_ID, product_id: int = PRODUCT_ID, 
                 interface: int = INTERFACE_NUM, packet_size: int = PACKET_SIZE, 
                 timeout_ms: int = 5000, 
                 # New Logging Arguments
                 log_file: Optional[str] = None, 
                 log_console: bool = True,
                 **kwargs):
        
        self.vendor_id = vendor_id
        self.product_id = product_id
        self.interface = interface
        self.packet_size = packet_size
        self.timeout_ms = timeout_ms
        self.pb = protobuf_module

        # --- Logging Setup ---
        # We create a logger unique to this instance to prevent collisions
        self.logger = logging.getLogger(f"LogicWeave-{id(self)}")
        self.logger.setLevel(logging.INFO)
        self.logger.handlers = [] # Reset handlers

        formatter = logging.Formatter('%(asctime)s | %(message)s', datefmt='%H:%M:%S')

        if log_console:
            console_handler = logging.StreamHandler(sys.stdout)
            console_handler.setFormatter(formatter)
            self.logger.addHandler(console_handler)

        if log_file:
            file_handler = logging.FileHandler(log_file)
            file_handler.setFormatter(formatter)
            self.logger.addHandler(file_handler)

        if transport is None:
            if sys.platform == 'emscripten': # 'emscripten' means we are in Pyodide
                transport = PyodideTransport
            else:
                transport = NativeUsbTransport

        # Setup Transport
        if isinstance(transport, type):
            self.transport = transport(vendor_id=self.vendor_id, 
                                     product_id=self.product_id, 
                                     interface=self.interface)
        else:
            self.transport = transport

        try:
            self.transport.open()
        except Exception as e:
            raise DeviceConnectionError(f"Connection failed: {e}")

    # --- New Logging Method ---
    def log(self, key: str, value: Any, level: int = logging.INFO):
        """
        Unified logging method.
        1. Updates Web GUI if 'update_gui' is available (injected in browser).
        2. Logs to console/file via standard logging, preserving structure.
        
        Using 'extra' allows structured loggers (like JSON formatters) to 
        access the raw 'output_key' and 'output_value' without parsing the message string.
        """
        try:
            update_gui(key, value)
        except:
            pass
        
        # B. Log to Python Logger (Console/File)
        # We pass key/value in 'extra' for better integration with logging backends
        self.logger.log(level, f"{key}: {value}", extra={'output_key': key, 'output_value': value})

    # --- Peripheral Factory Methods ---
    def uart(self, instance_num: int, tx_pin: int = DEFAULT_GPIO_PIN, rx_pin: int = DEFAULT_GPIO_PIN, baud_rate: int = 115200, name: str = "uart") -> 'UART':
        return UART(self, instance_num, tx_pin, rx_pin, baud_rate)

    def gpio(self, pin: int, name: str = "gpio") -> GPIO:
        return GPIO(self, pin, name)

    def i2c(self, instance_num: int, sda_pin: int = DEFAULT_GPIO_PIN, scl_pin: int = DEFAULT_GPIO_PIN, name: str = "i2c") -> I2C:
        return I2C(self, instance_num, sda_pin, scl_pin, name)

    def spi(self, instance_num: int, sclk_pin: int, mosi_pin: int, miso_pin: int = DEFAULT_GPIO_PIN, baud_rate: int = 1000000, default_cs_pin: Optional[int] = None, name: str = "spi") -> SPI:
        return SPI(self, instance_num, sclk_pin, mosi_pin, miso_pin, baud_rate, name, default_cs_pin)

    def _execute_transaction(self, specific_message_payload):
        app_message = self.pb.RequestMessage()
        
        field_name = None
        for field in app_message.DESCRIPTOR.fields:
            if field.containing_oneof and field.message_type == specific_message_payload.DESCRIPTOR:
                field_name = field.name
                break
        
        if not field_name:
            raise ValueError(f"Could not find a field in RequestMessage for: {type(specific_message_payload).__name__}.")
        
        getattr(app_message, field_name).CopyFrom(specific_message_payload)
        
        request_bytes = app_message.SerializeToString()
        length = len(request_bytes)
        
        MAX_PAYLOAD_SIZE = self.packet_size - 1
        if length > MAX_PAYLOAD_SIZE:
            raise ValueError(f"Message too large: {length} bytes.")

        length_prefix = struct.pack("B", length) 
        padding_needed = self.packet_size - (1 + length)
        padding = b'\x00' * padding_needed
        packet_to_send = length_prefix + request_bytes + padding

        try:
            self.transport.write(packet_to_send)
            full_response_bytes = self.transport.read(self.packet_size, 5000)
        except Exception as e:
            raise DeviceConnectionError(f"USB Transfer Error: {e}") from e
        
        if len(full_response_bytes) == 0:
            return self.pb.ResponseMessage() 

        response_length = full_response_bytes[0]
        response_bytes = full_response_bytes[1 : 1 + response_length]
        
        try:
            parsed_response = self.pb.ResponseMessage()
            parsed_response.ParseFromString(response_bytes)
            return parsed_response
        except Exception as e:
            raise DeviceFirmwareError(f"Client-side parse error: {e}")

    def _send_and_parse(self, request_payload, expected_response_field: str):
        response_app_msg = self._execute_transaction(request_payload)
        response_field = response_app_msg.WhichOneof("kind")
        if response_field == "error_response":
            raise DeviceFirmwareError(f"{response_app_msg.error_response.message}")
        if response_field != expected_response_field:
            raise DeviceResponseError(expected=expected_response_field, received=response_field)
        return getattr(response_app_msg, response_field)

    def close(self):
        if self.transport:
            self.transport.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    # --- High-Level API Methods ---
    def read_firmware_info(self) -> 'ProtobufModule.FirmwareInfoResponse':
        request = self.pb.FirmwareInfoRequest()
        return self._send_and_parse(request, "firmware_info_response")

    def write_bootloader_request(self):
        request = self.pb.UsbBootloaderRequest(val=1)
        self._send_and_parse(request, "usb_bootloader_response")

    def read_pin_function(self, gpio_pin):
        request = self.pb.GPIOReadFunctionRequest(gpio_pin=gpio_pin)
        return self._send_and_parse(request, "gpio_read_function_response")


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: logicweave.proto
# Protobuf Python Version: 6.33.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    1,
    '',
    'logicweave.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10logicweave.proto\x12\nlogicweave\"\xef\x0b\n\x0eRequestMessage\x12@\n\x15gpio_function_request\x18\x01 \x01(\x0b\x32\x1f.logicweave.GPIOFunctionRequestH\x00\x12G\n\x19\x62\x65gin_transaction_request\x18\x02 \x01(\x0b\x32\".logicweave.BeginTransationRequestH\x00\x12\x43\n\x17\x65nd_transaction_request\x18\x05 \x01(\x0b\x32 .logicweave.EndTransationRequestH\x00\x12:\n\x12gpio_write_request\x18\x03 \x01(\x0b\x32\x1c.logicweave.GPIOWriteRequestH\x00\x12\x42\n\x16usb_bootloader_request\x18\x04 \x01(\x0b\x32 .logicweave.UsbBootloaderRequestH\x00\x12\x38\n\x11gpio_read_request\x18\x06 \x01(\x0b\x32\x1b.logicweave.GPIOReadRequestH\x00\x12\x38\n\x11i2c_setup_request\x18\x0b \x01(\x0b\x32\x1b.logicweave.I2CSetupRequestH\x00\x12\x38\n\x11i2c_write_request\x18\r \x01(\x0b\x32\x1b.logicweave.I2CWriteRequestH\x00\x12\x36\n\x10i2c_read_request\x18\x0f \x01(\x0b\x32\x1a.logicweave.I2CReadRequestH\x00\x12\x38\n\x11spi_setup_request\x18\x11 \x01(\x0b\x32\x1b.logicweave.SPISetupRequestH\x00\x12\x36\n\x10spi_read_request\x18\x13 \x01(\x0b\x32\x1a.logicweave.SPIReadRequestH\x00\x12\x38\n\x11spi_write_request\x18\x15 \x01(\x0b\x32\x1b.logicweave.SPIWriteRequestH\x00\x12@\n\x15\x66irmware_info_request\x18$ \x01(\x0b\x32\x1f.logicweave.FirmwareInfoRequestH\x00\x12:\n\x12uart_setup_request\x18\x31 \x01(\x0b\x32\x1c.logicweave.UartSetupRequestH\x00\x12\x38\n\x11uart_read_request\x18\x33 \x01(\x0b\x32\x1b.logicweave.UartReadRequestH\x00\x12:\n\x12uart_write_request\x18\x35 \x01(\x0b\x32\x1c.logicweave.UartWriteRequestH\x00\x12?\n\x15gpio_pin_pull_request\x18\x37 \x01(\x0b\x32\x1e.logicweave.GpioPinPullRequestH\x00\x12\x36\n\x10\x61\x64\x63_read_request\x18\x39 \x01(\x0b\x32\x1a.logicweave.ADCReadRequestH\x00\x12J\n\x1bi2c_write_then_read_request\x18; \x01(\x0b\x32#.logicweave.I2CWriteThenReadRequestH\x00\x12\x38\n\x11pwm_setup_request\x18= \x01(\x0b\x32\x1b.logicweave.PWMSetupRequestH\x00\x12?\n\x15pwm_set_level_request\x18? \x01(\x0b\x32\x1e.logicweave.PWMSetLevelRequestH\x00\x12\x36\n\x10sleep_ms_request\x18\x41 \x01(\x0b\x32\x1a.logicweave.SleepMsRequestH\x00\x12I\n\x1agpio_read_function_request\x18\x43 \x01(\x0b\x32#.logicweave.GPIOReadFunctionRequestH\x00\x12<\n\x13stream_uart_request\x18O \x01(\x0b\x32\x1d.logicweave.StreamUartRequestH\x00\x42\x06\n\x04kind\"\x91\x0b\n\x0fResponseMessage\x12\x37\n\x1a\x62\x65gin_transaction_response\x18\x01 \x01(\x0b\x32\x11.logicweave.EmptyH\x00\x12\x33\n\x16\x65nd_transaction_respse\x18\x02 \x01(\x0b\x32\x11.logicweave.EmptyH\x00\x12\x33\n\x0e\x65rror_response\x18\n \x01(\x0b\x32\x19.logicweave.ErrorResponseH\x00\x12\x33\n\x16gpio_function_response\x18* \x01(\x0b\x32\x11.logicweave.EmptyH\x00\x12\x30\n\x13gpio_write_response\x18\x05 \x01(\x0b\x32\x11.logicweave.EmptyH\x00\x12\x34\n\x17usb_bootloader_response\x18\t \x01(\x0b\x32\x11.logicweave.EmptyH\x00\x12:\n\x12gpio_read_response\x18\x07 \x01(\x0b\x32\x1c.logicweave.GPIOReadResponseH\x00\x12/\n\x12i2c_setup_response\x18\x0c \x01(\x0b\x32\x11.logicweave.EmptyH\x00\x12/\n\x12i2c_write_response\x18\x0e \x01(\x0b\x32\x11.logicweave.EmptyH\x00\x12\x38\n\x11i2c_read_response\x18\x10 \x01(\x0b\x32\x1b.logicweave.I2CReadResponseH\x00\x12/\n\x12spi_setup_response\x18\x12 \x01(\x0b\x32\x11.logicweave.EmptyH\x00\x12\x38\n\x11spi_read_response\x18\x14 \x01(\x0b\x32\x1b.logicweave.SPIReadResponseH\x00\x12/\n\x12spi_write_response\x18\x16 \x01(\x0b\x32\x11.logicweave.EmptyH\x00\x12\x42\n\x16\x66irmware_info_response\x18% \x01(\x0b\x32 .logicweave.FirmwareInfoResponseH\x00\x12\x30\n\x13uart_setup_response\x18\x32 \x01(\x0b\x32\x11.logicweave.EmptyH\x00\x12:\n\x12uart_read_response\x18\x34 \x01(\x0b\x32\x1c.logicweave.UartReadResponseH\x00\x12\x30\n\x13uart_write_response\x18\x36 \x01(\x0b\x32\x11.logicweave.EmptyH\x00\x12\x33\n\x16gpio_pin_pull_response\x18\x38 \x01(\x0b\x32\x11.logicweave.EmptyH\x00\x12\x38\n\x11\x61\x64\x63_read_response\x18: \x01(\x0b\x32\x1b.logicweave.ADCReadResponseH\x00\x12L\n\x1ci2c_write_then_read_response\x18< \x01(\x0b\x32$.logicweave.I2CWriteThenReadResponseH\x00\x12/\n\x12pwm_setup_response\x18> \x01(\x0b\x32\x11.logicweave.EmptyH\x00\x12\x33\n\x16pwm_set_level_response\x18@ \x01(\x0b\x32\x11.logicweave.EmptyH\x00\x12.\n\x11sleep_ms_response\x18\x42 \x01(\x0b\x32\x11.logicweave.EmptyH\x00\x12K\n\x1bgpio_read_function_response\x18\x44 \x01(\x0b\x32$.logicweave.GPIOReadFunctionResponseH\x00\x12>\n\x14stream_uart_response\x18Q \x01(\x0b\x32\x1e.logicweave.StreamUartResponseH\x00\x42\x06\n\x04kind\"m\n\x11StreamUartRequest\x12\x0f\n\x07\x65nabled\x18\x01 \x01(\x08\x12\x11\n\tbaud_rate\x18\x02 \x01(\r\x12\x0e\n\x06tx_pin\x18\x04 \x01(\r\x12\x0e\n\x06rx_pin\x18\x05 \x01(\r\x12\x14\n\x0cinstance_num\x18\x06 \x01(\r\"\"\n\x12StreamUartResponse\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\"\'\n\x16\x42\x65ginTransationRequest\x12\r\n\x05label\x18\x01 \x01(\t\"6\n\x14\x45ndTransationRequest\x12\r\n\x05label\x18\x01 \x01(\t\x12\x0f\n\x07success\x18\x02 \x01(\x08\"+\n\x17GPIOReadFunctionRequest\x12\x10\n\x08gpio_pin\x18\x01 \x01(\r\"F\n\x18GPIOReadFunctionResponse\x12*\n\x08\x66unction\x18\x01 \x01(\x0e\x32\x18.logicweave.GpioFunction\"#\n\x14UsbBootloaderRequest\x12\x0b\n\x03val\x18\x01 \x01(\r\"\x17\n\x05\x45mpty\x12\x0e\n\x06status\x18\x01 \x01(\r\"\x1e\n\x0b\x45\x63hoMessage\x12\x0f\n\x07message\x18\x01 \x01(\t\"#\n\x0fGPIOReadRequest\x12\x10\n\x08gpio_pin\x18\x01 \x01(\r\"n\n\x0fPWMSetupRequest\x12\x10\n\x08gpio_pin\x18\x01 \x01(\r\x12\x0c\n\x04wrap\x18\x02 \x01(\r\x12\x15\n\rclock_div_int\x18\x03 \x01(\r\x12\x16\n\x0e\x63lock_div_frac\x18\x04 \x01(\r\x12\x0c\n\x04name\x18\x05 \x01(\t\"5\n\x12PWMSetLevelRequest\x12\x10\n\x08gpio_pin\x18\x01 \x01(\r\x12\r\n\x05level\x18\x02 \x01(\r\"\"\n\x0e\x41\x44\x43ReadRequest\x12\x10\n\x08gpio_pin\x18\x01 \x01(\r\"!\n\x0f\x41\x44\x43ReadResponse\x12\x0e\n\x06sample\x18\x01 \x01(\r\"!\n\x10GPIOReadResponse\x12\r\n\x05state\x18\x01 \x01(\x08\"3\n\x10GPIOWriteRequest\x12\x10\n\x08gpio_pin\x18\x01 \x01(\r\x12\r\n\x05state\x18\x02 \x01(\x08\"~\n\x0fSPISetupRequest\x12\x10\n\x08sclk_pin\x18\x02 \x01(\r\x12\x10\n\x08mosi_pin\x18\x03 \x01(\r\x12\x10\n\x08miso_pin\x18\x04 \x01(\r\x12\x14\n\x0cinstance_num\x18\x05 \x01(\r\x12\x11\n\tbaud_rate\x18\x06 \x01(\r\x12\x0c\n\x04name\x18\x07 \x01(\t\"E\n\x0fSPIWriteRequest\x12\x14\n\x0cinstance_num\x18\x01 \x01(\r\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x12\x0e\n\x06\x63s_pin\x18\x03 \x01(\r\"X\n\x0eSPIReadRequest\x12\x14\n\x0cinstance_num\x18\x01 \x01(\r\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\r\x12\x0e\n\x06\x63s_pin\x18\x03 \x01(\r\x12\x12\n\nbyte_count\x18\x04 \x01(\r\"\x1f\n\x0fSPIReadResponse\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\"O\n\x12GpioPinPullRequest\x12\x10\n\x08gpio_pin\x18\x01 \x01(\r\x12\'\n\x05state\x18\x02 \x01(\x0e\x32\x18.logicweave.PinPullState\"W\n\x0fI2CSetupRequest\x12\x0f\n\x07sda_pin\x18\x01 \x01(\r\x12\x0f\n\x07scl_pin\x18\x02 \x01(\r\x12\x14\n\x0cinstance_num\x18\x03 \x01(\r\x12\x0c\n\x04name\x18\x04 \x01(\t\"M\n\x0fI2CWriteRequest\x12\x14\n\x0cinstance_num\x18\x01 \x01(\r\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x12\x16\n\x0e\x64\x65vice_address\x18\x03 \x01(\r\"R\n\x0eI2CReadRequest\x12\x14\n\x0cinstance_num\x18\x01 \x01(\r\x12\x16\n\x0e\x64\x65vice_address\x18\x02 \x01(\r\x12\x12\n\nbyte_count\x18\x03 \x01(\r\"i\n\x17I2CWriteThenReadRequest\x12\x14\n\x0cinstance_num\x18\x01 \x01(\r\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x12\x16\n\x0e\x64\x65vice_address\x18\x03 \x01(\r\x12\x12\n\nbyte_count\x18\x04 \x01(\r\"(\n\x18I2CWriteThenReadResponse\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\"\x1f\n\x0fI2CReadResponse\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\"a\n\x13GPIOFunctionRequest\x12\x10\n\x08gpio_pin\x18\x01 \x01(\r\x12*\n\x08\x66unction\x18\x02 \x01(\x0e\x32\x18.logicweave.GpioFunction\x12\x0c\n\x04name\x18\x03 \x01(\t\"i\n\x10UartSetupRequest\x12\x11\n\tbaud_rate\x18\x01 \x01(\r\x12\x0e\n\x06tx_pin\x18\x02 \x01(\r\x12\x0e\n\x06rx_pin\x18\x03 \x01(\r\x12\x14\n\x0cinstance_num\x18\x04 \x01(\r\x12\x0c\n\x04name\x18\x05 \x01(\t\"J\n\x10UartWriteRequest\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x12\x12\n\ntimeout_ms\x18\x02 \x01(\r\x12\x14\n\x0cinstance_num\x18\x03 \x01(\r\"O\n\x0fUartReadRequest\x12\x12\n\nbyte_count\x18\x01 \x01(\r\x12\x12\n\ntimeout_ms\x18\x02 \x01(\r\x12\x14\n\x0cinstance_num\x18\x03 \x01(\r\" \n\x10UartReadResponse\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\"\x15\n\x13\x46irmwareInfoRequest\"b\n\x14\x46irmwareInfoResponse\x12\"\n\x06\x64\x65vice\x18\x01 \x01(\x0e\x32\x12.logicweave.Device\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x15\n\rserial_number\x18\x03 \x01(\t\"\"\n\x0eSleepMsRequest\x12\x10\n\x08sleep_ms\x18\x01 \x01(\r\" \n\rErrorResponse\x12\x0f\n\x07message\x18\x01 \x01(\t*\x85\x01\n\x0cGpioFunction\x12\x07\n\x03xip\x10\x00\x12\x08\n\x04hstx\x10\x01\x12\x07\n\x03spi\x10\x02\x12\x08\n\x04uart\x10\x03\x12\x07\n\x03i2c\x10\x04\x12\x07\n\x03pwm\x10\x05\x12\n\n\x06sio_in\x10\x06\x12\x0b\n\x07sio_out\x10\x07\x12\x07\n\x03pio\x10\x08\x12\x08\n\x04gpck\x10\t\x12\x07\n\x03usb\x10\n\x12\x08\n\x04none\x10\x0b*+\n\x0b\x42\x61nkVoltage\x12\x08\n\x04V1P8\x10\x00\x12\x08\n\x04V3P3\x10\x01\x12\x08\n\x04V5P0\x10\x02*2\n\x0cPinPullState\x12\x08\n\x04None\x10\x00\x12\n\n\x06PullUp\x10\x01\x12\x0c\n\x08PullDown\x10\x02*:\n\x06\x44\x65vice\x12\n\n\x06rp2040\x10\x00\x12\n\n\x06rp2350\x10\x01\x12\x0b\n\x07rp2350b\x10\x02\x12\x0b\n\x07lw_core\x10\x03\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'logicweave_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_GPIOFUNCTION']._serialized_start=5228
  _globals['_GPIOFUNCTION']._serialized_end=5361
  _globals['_BANKVOLTAGE']._serialized_start=5363
  _globals['_BANKVOLTAGE']._serialized_end=5406
  _globals['_PINPULLSTATE']._serialized_start=5408
  _globals['_PINPULLSTATE']._serialized_end=5458
  _globals['_DEVICE']._serialized_start=5460
  _globals['_DEVICE']._serialized_end=5518
  _globals['_REQUESTMESSAGE']._serialized_start=33
  _globals['_REQUESTMESSAGE']._serialized_end=1552
  _globals['_RESPONSEMESSAGE']._serialized_start=1555
  _globals['_RESPONSEMESSAGE']._serialized_end=2980
  _globals['_STREAMUARTREQUEST']._serialized_start=2982
  _globals['_STREAMUARTREQUEST']._serialized_end=3091
  _globals['_STREAMUARTRESPONSE']._serialized_start=3093
  _globals['_STREAMUARTRESPONSE']._serialized_end=3127
  _globals['_BEGINTRANSATIONREQUEST']._serialized_start=3129
  _globals['_BEGINTRANSATIONREQUEST']._serialized_end=3168
  _globals['_ENDTRANSATIONREQUEST']._serialized_start=3170
  _globals['_ENDTRANSATIONREQUEST']._serialized_end=3224
  _globals['_GPIOREADFUNCTIONREQUEST']._serialized_start=3226
  _globals['_GPIOREADFUNCTIONREQUEST']._serialized_end=3269
  _globals['_GPIOREADFUNCTIONRESPONSE']._serialized_start=3271
  _globals['_GPIOREADFUNCTIONRESPONSE']._serialized_end=3341
  _globals['_USBBOOTLOADERREQUEST']._serialized_start=3343
  _globals['_USBBOOTLOADERREQUEST']._serialized_end=3378
  _globals['_EMPTY']._serialized_start=3380
  _globals['_EMPTY']._serialized_end=3403
  _globals['_ECHOMESSAGE']._serialized_start=3405
  _globals['_ECHOMESSAGE']._serialized_end=3435
  _globals['_GPIOREADREQUEST']._serialized_start=3437
  _globals['_GPIOREADREQUEST']._serialized_end=3472
  _globals['_PWMSETUPREQUEST']._serialized_start=3474
  _globals['_PWMSETUPREQUEST']._serialized_end=3584
  _globals['_PWMSETLEVELREQUEST']._serialized_start=3586
  _globals['_PWMSETLEVELREQUEST']._serialized_end=3639
  _globals['_ADCREADREQUEST']._serialized_start=3641
  _globals['_ADCREADREQUEST']._serialized_end=3675
  _globals['_ADCREADRESPONSE']._serialized_start=3677
  _globals['_ADCREADRESPONSE']._serialized_end=3710
  _globals['_GPIOREADRESPONSE']._serialized_start=3712
  _globals['_GPIOREADRESPONSE']._serialized_end=3745
  _globals['_GPIOWRITEREQUEST']._serialized_start=3747
  _globals['_GPIOWRITEREQUEST']._serialized_end=3798
  _globals['_SPISETUPREQUEST']._serialized_start=3800
  _globals['_SPISETUPREQUEST']._serialized_end=3926
  _globals['_SPIWRITEREQUEST']._serialized_start=3928
  _globals['_SPIWRITEREQUEST']._serialized_end=3997
  _globals['_SPIREADREQUEST']._serialized_start=3999
  _globals['_SPIREADREQUEST']._serialized_end=4087
  _globals['_SPIREADRESPONSE']._serialized_start=4089
  _globals['_SPIREADRESPONSE']._serialized_end=4120
  _globals['_GPIOPINPULLREQUEST']._serialized_start=4122
  _globals['_GPIOPINPULLREQUEST']._serialized_end=4201
  _globals['_I2CSETUPREQUEST']._serialized_start=4203
  _globals['_I2CSETUPREQUEST']._serialized_end=4290
  _globals['_I2CWRITEREQUEST']._serialized_start=4292
  _globals['_I2CWRITEREQUEST']._serialized_end=4369
  _globals['_I2CREADREQUEST']._serialized_start=4371
  _globals['_I2CREADREQUEST']._serialized_end=4453
  _globals['_I2CWRITETHENREADREQUEST']._serialized_start=4455
  _globals['_I2CWRITETHENREADREQUEST']._serialized_end=4560
  _globals['_I2CWRITETHENREADRESPONSE']._serialized_start=4562
  _globals['_I2CWRITETHENREADRESPONSE']._serialized_end=4602
  _globals['_I2CREADRESPONSE']._serialized_start=4604
  _globals['_I2CREADRESPONSE']._serialized_end=4635
  _globals['_GPIOFUNCTIONREQUEST']._serialized_start=4637
  _globals['_GPIOFUNCTIONREQUEST']._serialized_end=4734
  _globals['_UARTSETUPREQUEST']._serialized_start=4736
  _globals['_UARTSETUPREQUEST']._serialized_end=4841
  _globals['_UARTWRITEREQUEST']._serialized_start=4843
  _globals['_UARTWRITEREQUEST']._serialized_end=4917
  _globals['_UARTREADREQUEST']._serialized_start=4919
  _globals['_UARTREADREQUEST']._serialized_end=4998
  _globals['_UARTREADRESPONSE']._serialized_start=5000
  _globals['_UARTREADRESPONSE']._serialized_end=5032
  _globals['_FIRMWAREINFOREQUEST']._serialized_start=5034
  _globals['_FIRMWAREINFOREQUEST']._serialized_end=5055
  _globals['_FIRMWAREINFORESPONSE']._serialized_start=5057
  _globals['_FIRMWAREINFORESPONSE']._serialized_end=5155
  _globals['_SLEEPMSREQUEST']._serialized_start=5157
  _globals['_SLEEPMSREQUEST']._serialized_end=5191
  _globals['_ERRORRESPONSE']._serialized_start=5193
  _globals['_ERRORRESPONSE']._serialized_end=5225
# @@protoc_insertion_point(module_scope)



# DEPRICIATED, USING WEBWORKERS INSTEAD

# web.py
import struct
import asyncio
import logging  # <--- Added logging import
from typing import Optional, Any
import sys

# Pyodide specific import (only available in browser)
try:
    import js
except ImportError:
    js = None

import LogicWeave.proto_gen.logicweave_pb2 as lw_pb2
from .transports import Transport, VENDOR_ID, PRODUCT_ID, INTERFACE_NUM, PACKET_SIZE
from .exceptions import DeviceConnectionError, DeviceFirmwareError, DeviceResponseError
# Import the base LogicWeave class to inherit constants/structure if needed, 
# though we largely redefine behaviors for async.
from .logicweave import LogicWeave as SyncLogicWeave

ProtobufModule = Any

# ==========================================
# 1. Async Web Transport
# ==========================================
class WebUSBTransport(Transport):
    """
    Adapts the browser's navigator.usb via the global 'window.usbBridge' 
    object defined in your HTML.
    """
    def __init__(self, js_interface_name="usbBridge", **kwargs):
        if not js:
            raise RuntimeError("WebUSBTransport can only run in a Pyodide/Browser environment.")
        self.js = js
        self.bridge = getattr(js, js_interface_name)

    async def ensure_open(self):
        if not self.bridge.isConnected():
            raise DeviceConnectionError("WebUSB device is not connected in the browser.")

    # In web, 'open' is handled by the UI permission prompt, not Python
    def open(self):
        pass

    async def write(self, data: bytes):
        # Convert Python bytes to JS Uint8Array
        js_array = self.js.Uint8Array.new(data)
        await self.bridge.writePacket(js_array)

    async def read(self, length: int, timeout_ms: int) -> bytes:
        # Returns a JS Uint8Array, convert back to bytes
        response_js = await self.bridge.readPacket(length, timeout_ms)
        return bytes(response_js.to_py())

    async def close(self):
        # Browser handles connection lifecycle mostly
        pass


# ==========================================
# 2. Async Base Peripheral
# ==========================================
class _AsyncBasePeripheral:
    """Async version of _BasePeripheral."""
    def __init__(self, controller: 'WebLogicWeave'):
        self._controller = controller
        self.pb: ProtobufModule = controller.pb

    async def _build_and_execute(self, request_class, expected_response_field: str, **kwargs):
        request_payload = request_class(**kwargs)
        return await self._controller._send_and_parse(request_payload, expected_response_field)


# ==========================================
# 3. Async Peripheral Implementations
# ==========================================

class AsyncUART(_AsyncBasePeripheral):
    def __init__(self, controller: 'WebLogicWeave', instance_num: int, tx_pin: int, rx_pin: int, baud_rate: int):
        super().__init__(controller)
        self._instance_num = instance_num
        self.tx_pin = tx_pin
        self.rx_pin = rx_pin
        self.baud_rate = baud_rate
        # Async setup task
        asyncio.create_task(self._setup())

    async def _setup(self):
        await self._build_and_execute(self.pb.UartSetupRequest, "uart_setup_response", 
                                      instance_num=self._instance_num, tx_pin=self.tx_pin, 
                                      rx_pin=self.rx_pin, baud_rate=self.baud_rate)

    async def write(self, data: bytes, timeout_ms: int = 1000):
        await self._build_and_execute(self.pb.UartWriteRequest, "uart_write_response", 
                                      instance_num=self._instance_num, data=data, 
                                      timeout_ms=timeout_ms)

    async def read(self, byte_count: int, timeout_ms: int = 1000) -> bytes:
        response = await self._build_and_execute(self.pb.UartReadRequest, "uart_read_response", 
                                                 instance_num=self._instance_num, 
                                                 byte_count=byte_count, timeout_ms=timeout_ms)
        return response.data

    def __repr__(self):
        return f"<AsyncUART instance={self._instance_num} tx={self.tx_pin} rx={self.rx_pin}>"


class AsyncGPIO(_AsyncBasePeripheral):
    MAX_ADC_COUNT = 4095
    V_REF = 3.3

    def __init__(self, controller: 'WebLogicWeave', pin: int, name: Optional[str] = "gpio"):
        super().__init__(controller)
        self.pin = pin
        self.pull = None
        self.name = name

    async def set_function(self, mode: int):
        await self._build_and_execute(self.pb.GPIOFunctionRequest, "gpio_function_response", 
                                      gpio_pin=self.pin, function=mode, name=self.name)

    async def set_pull(self, state: int):
        await self._build_and_execute(self.pb.GpioPinPullRequest, "gpio_pin_pull_response", 
                                      gpio_pin=self.pin, state=state)
        self.pull = state

    async def write(self, state: bool):
        # Async check for current function
        current_func = await self._controller.read_pin_function(self.pin)
        if current_func != self.pb.GpioFunction.sio_out:
             await self.set_function(self.pb.GpioFunction.sio_out)
        
        await self._build_and_execute(self.pb.GPIOWriteRequest, "gpio_write_response", 
                                      gpio_pin=self.pin, state=state)

    async def read(self) -> bool:
        current_func = await self._controller.read_pin_function(self.pin)
        if current_func != self.pb.GpioFunction.sio_in:
             await self.set_function(self.pb.GpioFunction.sio_in)
             
        response = await self._build_and_execute(self.pb.GPIOReadRequest, "gpio_read_response", 
                                                 gpio_pin=self.pin)
        return response.state

    async def setup_pwm(self, wrap, clock_div_int=0, clock_div_frac=0):
        await self._build_and_execute(self.pb.PWMSetupRequest, "pwm_setup_response", 
                                      gpio_pin=self.pin, wrap=wrap, 
                                      clock_div_int=clock_div_int, 
                                      clock_div_frac=clock_div_frac, name=self.name)

    async def set_pwm_level(self, level):
        await self._build_and_execute(self.pb.PWMSetLevelRequest, "pwm_set_level_response", 
                                      gpio_pin=self.pin, level=level)

    async def read_adc(self) -> float:
        response = await self._build_and_execute(self.pb.ADCReadRequest, "adc_read_response", 
                                                 gpio_pin=self.pin)
        return (response.sample / self.MAX_ADC_COUNT) * self.V_REF

    def __repr__(self):
        return f"<AsyncGPIO pin={self.pin}>"


class AsyncI2C(_AsyncBasePeripheral):
    def __init__(self, controller: 'WebLogicWeave', instance_num: int, sda_pin: int, scl_pin: int, name: Optional[str] = "i2c"):
        super().__init__(controller)
        self._instance_num = instance_num
        self.sda_pin = sda_pin
        self.scl_pin = scl_pin
        self.name = name
        # Async setup task
        asyncio.create_task(self._setup())

    async def _setup(self):
        await self._build_and_execute(self.pb.I2CSetupRequest, "i2c_setup_response", 
                                      instance_num=self._instance_num, sda_pin=self.sda_pin, 
                                      scl_pin=self.scl_pin, name=self.name)

    async def write(self, device_address: int, data: bytes):
        await self._build_and_execute(self.pb.I2CWriteRequest, "i2c_write_response", 
                                      instance_num=self._instance_num, 
                                      device_address=device_address, data=data)

    async def write_then_read(self, device_address: int, data: bytes, byte_count: int) -> bytes:
        response = await self._build_and_execute(self.pb.I2CWriteThenReadRequest, "i2c_write_then_read_response", 
                                                 instance_num=self._instance_num, 
                                                 device_address=device_address, data=data, 
                                                 byte_count=byte_count)
        return response.data

    async def read(self, device_address: int, byte_count: int) -> bytes:
        response = await self._build_and_execute(self.pb.I2CReadRequest, "i2c_read_response", 
                                                 instance_num=self._instance_num, 
                                                 device_address=device_address, 
                                                 byte_count=byte_count)
        return response.data

    def __repr__(self):
        return f"<AsyncI2C instance={self._instance_num} sda={self.sda_pin} scl={self.scl_pin}>"


class AsyncSPI(_AsyncBasePeripheral):
    def __init__(self, controller: 'WebLogicWeave', instance_num: int, sclk_pin: int, mosi_pin: int, miso_pin: int, baud_rate: int, name: Optional[str] = "spi", default_cs_pin: Optional[int] = None):
        super().__init__(controller)
        self._instance_num = instance_num
        self.sclk_pin = sclk_pin
        self.mosi_pin = mosi_pin
        self.miso_pin = miso_pin
        self.baud_rate = baud_rate
        self._default_cs_pin = default_cs_pin
        self.name = name
        # Async setup task
        asyncio.create_task(self._setup())

    async def _setup(self):
        await self._build_and_execute(self.pb.SPISetupRequest, "spi_setup_response", 
                                      instance_num=self._instance_num, sclk_pin=self.sclk_pin, 
                                      mosi_pin=self.mosi_pin, miso_pin=self.miso_pin, 
                                      baud_rate=self.baud_rate, name=self.name)

    def _get_cs_pin(self, cs_pin_override: Optional[int]) -> int:
        # This helper remains sync as it's just logic
        active_cs_pin = cs_pin_override if cs_pin_override is not None else self._default_cs_pin
        if active_cs_pin is None: 
            raise ValueError("A Chip Select (CS) pin must be provided.")
        return active_cs_pin

    async def write(self, data: bytes, cs_pin: Optional[int] = None):
        await self._build_and_execute(self.pb.SPIWriteRequest, "spi_write_response", 
                                      instance_num=self._instance_num, data=data, 
                                      cs_pin=self._get_cs_pin(cs_pin))

    async def read(self, byte_count: int, cs_pin: Optional[int] = None, data_to_send: int = 0) -> bytes:
        response = await self._build_and_execute(self.pb.SPIReadRequest, "spi_read_response", 
                                                 instance_num=self._instance_num, 
                                                 data=data_to_send, 
                                                 cs_pin=self._get_cs_pin(cs_pin), 
                                                 byte_count=byte_count)
        return response.data

    def __repr__(self):
        return f"<AsyncSPI instance={self._instance_num}>"


# ==========================================
# 4. Main Async Controller Class
# ==========================================
class WebLogicWeave(SyncLogicWeave):
    """
    An Async/Await version of the LogicWeave controller designed for the Web/Pyodide.
    It inherits from SyncLogicWeave to keep constant/init signature compatibility,
    but overrides all operational methods to be async.
    """
    def __init__(self, transport=None, protobuf_module=lw_pb2, 
                 vendor_id=VENDOR_ID, product_id=PRODUCT_ID, 
                 interface=INTERFACE_NUM, packet_size=PACKET_SIZE, 
                 timeout_ms=5000, 
                 # New Logging Arguments
                 log_file: Optional[str] = None, 
                 log_console: bool = True,
                 **kwargs):
        
        self.vendor_id = vendor_id
        self.product_id = product_id
        self.interface = interface
        self.packet_size = packet_size
        self.timeout_ms = timeout_ms
        self.pb = protobuf_module

        # --- Logging Setup (Ported from LogicWeave) ---
        # We create a logger unique to this instance to prevent collisions
        self.logger = logging.getLogger(f"LogicWeave-{id(self)}")
        self.logger.setLevel(logging.INFO)
        self.logger.handlers = [] # Reset handlers

        formatter = logging.Formatter('%(asctime)s | %(message)s', datefmt='%H:%M:%S')

        if log_console:
            console_handler = logging.StreamHandler(sys.stdout)
            console_handler.setFormatter(formatter)
            self.logger.addHandler(console_handler)

        if log_file:
            # Note: In Pyodide, this writes to the in-memory Emscripten filesystem
            file_handler = logging.FileHandler(log_file)
            file_handler.setFormatter(formatter)
            self.logger.addHandler(file_handler)

        # Force WebUSBTransport if not provided
        if transport is None:
            self.transport = WebUSBTransport(vendor_id=self.vendor_id, 
                                             product_id=self.product_id)
        else:
            self.transport = transport

    # --- New Logging Method (Ported from LogicWeave) ---
    def log(self, key: str, value: Any, level: int = logging.INFO):
        """
        Unified logging method.
        1. Updates Web GUI if 'update_gui' is available.
        2. Logs to console/file via standard logging, preserving structure.
        """
        try:
            # Check for global update_gui function (common in Pyodide/GUI implementations)
            update_gui(key, value)
        except:
            pass
        
        # Log to Python Logger (Console/File)
        self.logger.log(level, f"{key}: {value}", extra={'output_key': key, 'output_value': value})

    # --- Async Context Manager ---
    async def __aenter__(self):
        await self.transport.ensure_open()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def close(self):
        await self.transport.close()

    # --- Overridden Factory Methods (Return Async Wrappers) ---
    def uart(self, instance_num: int, tx_pin: int, rx_pin: int, baud_rate: int = 115200, name: str = "uart") -> AsyncUART:
        return AsyncUART(self, instance_num, tx_pin, rx_pin, baud_rate)

    def gpio(self, pin: int, name: str = "gpio") -> AsyncGPIO:
        return AsyncGPIO(self, pin, name)

    def i2c(self, instance_num: int, sda_pin: int, scl_pin: int, name: str = "i2c") -> AsyncI2C:
        return AsyncI2C(self, instance_num, sda_pin, scl_pin, name)

    def spi(self, instance_num: int, sclk_pin: int, mosi_pin: int, miso_pin: int, baud_rate: int = 1000000, default_cs_pin: Optional[int] = None, name: str = "spi") -> AsyncSPI:
        return AsyncSPI(self, instance_num, sclk_pin, mosi_pin, miso_pin, baud_rate, name, default_cs_pin)

    # --- Core Async Execution Logic ---
    async def _execute_transaction(self, specific_message_payload):
        app_message = self.pb.RequestMessage()
        
        field_name = None
        for field in app_message.DESCRIPTOR.fields:
            if field.containing_oneof and field.message_type == specific_message_payload.DESCRIPTOR:
                field_name = field.name
                break
        
        if not field_name:
            raise ValueError(f"Could not find a field in RequestMessage for: {type(specific_message_payload).__name__}.")
        
        getattr(app_message, field_name).CopyFrom(specific_message_payload)
        
        request_bytes = app_message.SerializeToString()
        length = len(request_bytes)
        
        MAX_PAYLOAD_SIZE = self.packet_size - 1
        if length > MAX_PAYLOAD_SIZE:
            raise ValueError(f"Message too large: {length} bytes.")

        length_prefix = struct.pack("B", length) 
        padding_needed = self.packet_size - (1 + length)
        padding = b'\x00' * padding_needed
        packet_to_send = length_prefix + request_bytes + padding

        try:
            # AWAIT HERE: Transport write
            await self.transport.write(packet_to_send)
            # AWAIT HERE: Transport read
            full_response_bytes = await self.transport.read(self.packet_size, 5000)
        except Exception as e:
            raise DeviceConnectionError(f"USB Transfer Error: {e}") from e
        
        if len(full_response_bytes) == 0:
            return self.pb.ResponseMessage() 

        response_length = full_response_bytes[0]
        response_bytes = full_response_bytes[1 : 1 + response_length]
        
        try:
            parsed_response = self.pb.ResponseMessage()
            parsed_response.ParseFromString(response_bytes)
            return parsed_response
        except Exception as e:
            raise DeviceFirmwareError(f"Client-side parse error: {e}")

    async def _send_and_parse(self, request_payload, expected_response_field: str):
        # AWAIT HERE
        response_app_msg = await self._execute_transaction(request_payload)
        
        response_field = response_app_msg.WhichOneof("kind")
        if response_field == "error_response":
             raise DeviceFirmwareError(f"{response_app_msg.error_response.message}")
        if response_field != expected_response_field:
             raise DeviceResponseError(expected=expected_response_field, received=response_field)
        return getattr(response_app_msg, response_field)

    # --- High-Level Async API Methods ---
    async def read_firmware_info(self) -> 'ProtobufModule.FirmwareInfoResponse':
        request = self.pb.FirmwareInfoRequest()
        return await self._send_and_parse(request, "firmware_info_response")

    async def write_bootloader_request(self):
        request = self.pb.UsbBootloaderRequest(val=1)
        await self._send_and_parse(request, "usb_bootloader_response")

    async def read_pin_function(self, gpio_pin):
        request = self.pb.GPIOReadFunctionRequest(gpio_pin=gpio_pin)
        return await self._send_and_parse(request, "gpio_read_function_response")
    
import LogicWeave.proto_gen.logicweave_core_pb2 as logicweave_core_pb2

# Inherit from the Async Web Controller you defined previously
class WebLogicWeaveCore(WebLogicWeave):
    """
    Async implementation of LogicWeaveCore for Browser/Pyodide.
    Inherits from WebLogicWeave to get the Async WebUSB transport
    and creates async wrappers for the Core specific functionality.
    """
    def __init__(self, *args, **kwargs):
        # Inject the extended protobuf definitions (core_pb2) 
        # so self.pb has the voltage/resistance/etc message definitions
        kwargs['protobuf_module'] = logicweave_core_pb2
        super().__init__(*args, **kwargs)

    async def read_voltage(self):
        request = self.pb.ReadVoltageRequest()
        return await self._send_and_parse(request, "read_voltage_response")

    async def read_resistance(self):
        request = self.pb.ReadResistanceRequest()
        return await self._send_and_parse(request, "read_resistance_response")
    
    async def read_pd(self):
        request = self.pb.ReadPDRequest()
        return await self._send_and_parse(request, "read_pd_response")
    
    async def set_psu_output(self, channel, state):
        request = self.pb.SetPSUOutputRequest(channel=channel, state=state)
        return await self._send_and_parse(request, "set_psu_output_response")
    
    async def read_power_monitor(self):
        request = self.pb.ReadPowerMonitorRequest()
        return await self._send_and_parse(request, "read_power_monitor_response")
    
    async def configure_psu(self, channel, voltage, current_limit):
        request = self.pb.ConfigurePSURequest(
            channel=channel, 
            voltage=voltage, 
            current_limit=current_limit
        )
        return await self._send_and_parse(request, "configure_psu_response")
    
    async def cal_probes(self):
        request = self.pb.ZeroProbesRequest()
        return await self._send_and_parse(request, "zero_probes_response")

    async def read_calibration_data(self):
        request = self.pb.ReadCalibrationDataRequest()
        return await self._send_and_parse(request, "read_calibration_data_response")


from .logicweave import LogicWeave, GPIO, UART, I2C, SPI, GpioFunction
from .web import WebLogicWeave, WebLogicWeaveCore
from .logicweave_core import LogicWeaveCore
from .proto_gen import logicweave_pb2, logicweave_core_pb2


import time
import platform
from typing import Protocol, Any

# --- CONSTANTS ---
# Kept here because the Transport layer needs them to find the device.
VENDOR_ID = 0x2E8A
PRODUCT_ID = 0x000a
INTERFACE_NUM = 1
EP_IN_ADDR = 0x84
EP_OUT_ADDR = 0x05
PACKET_SIZE = 64

class Transport(Protocol):
    def open(self) -> None: ...
    def write(self, data: bytes) -> None: ...
    def read(self, length: int, timeout_ms: int) -> bytes: ...
    def close(self) -> None: ...

class NativeUsbTransport:
    """Implementation using PyUSB for running locally on Windows/Linux/Mac."""
    def __init__(self, vendor_id=VENDOR_ID, product_id=PRODUCT_ID, interface=0):
        # Local imports so this class doesn't crash on the web
        import usb.core
        import usb.util
        self.usb_core = usb.core
        self.usb_util = usb.util
        self.platform = platform
        
        self.vendor_id = vendor_id
        self.product_id = product_id
        self.interface = interface
        self.dev = None
        self.ep_out = None
        self.ep_in = None
        self.kernel_driver_detached = False

    def open(self):
        self.dev = self.usb_core.find(idVendor=self.vendor_id, idProduct=self.product_id)
        if self.dev is None:
            raise Exception("Device not found")

        # Kernel driver detachment logic
        if self.platform.system() != "Windows":
            try:
                if self.dev.is_kernel_driver_active(self.interface):
                    self.dev.detach_kernel_driver(self.interface)
                    self.kernel_driver_detached = True
            except (NotImplementedError, self.usb_core.USBError):
                pass

        self.dev.set_configuration()
        cfg = self.dev.get_active_configuration()
        intf = cfg[(self.interface, 0)]

        self.ep_out = self.usb_util.find_descriptor(
            intf,
            custom_match=lambda e: e.bEndpointAddress == EP_OUT_ADDR
        )
        
        self.ep_in = self.usb_util.find_descriptor(
            intf,
            custom_match=lambda e: e.bEndpointAddress == EP_IN_ADDR
        )

        if self.ep_out is None:
             raise Exception(f"Endpoint OUT (0x{EP_OUT_ADDR:02x}) not found")
        if self.ep_in is None:
             raise Exception(f"Endpoint IN (0x{EP_IN_ADDR:02x}) not found")

    def write(self, data: bytes):
        if not self.ep_out: raise Exception("Not connected")
        self.ep_out.write(data)

    def read(self, length: int, timeout_ms: int) -> bytes:
        if not self.ep_in: raise Exception("Not connected")
        return bytes(self.ep_in.read(length, timeout=timeout_ms))

    def close(self):
        if self.dev:
            try:
                self.usb_util.release_interface(self.dev, self.interface)
                if self.kernel_driver_detached:
                    self.dev.attach_kernel_driver(self.interface)
                self.usb_util.dispose_resources(self.dev)
            except:
                pass

class PyodideTransport:
    """
    WebUSB Transport for running inside the Browser via Pyodide.
    Uses pyodide.ffi.run_sync to bridge async JS calls to sync Python.
    """
    def __init__(self, vendor_id=VENDOR_ID, product_id=PRODUCT_ID, interface=INTERFACE_NUM):
        self.vendor_id = vendor_id
        self.product_id = product_id
        self.interface = interface
        self.js_dev = None

    def open(self):
        import js
        from pyodide.ffi import run_sync
        
        # 1. Check if worker.js set the device globally
        if not hasattr(js, 'py_device') or not js.py_device:
            raise Exception("No authorized USB device found in browser context.")
        
        # 2. Define the async JS steps
        async def _async_open():
            dev = js.py_device
            await dev.open()
            await dev.selectConfiguration(1)
            await dev.claimInterface(self.interface)
            return dev

        # 3. Execute synchronously
        try:
            self.js_dev = run_sync(_async_open())
        except Exception as e:
            raise Exception(f"Failed to open WebUSB device: {e}")

    def write(self, data: bytes):
        if not self.js_dev:
            raise Exception("Device not open")
        
        import js
        from pyodide.ffi import run_sync

        # Convert Python bytes to JS Uint8Array
        js_data = js.Uint8Array.new(data)

        async def _async_write():
            # WebUSB transferOut(endpointNumber, data)
            # EP_OUT_ADDR is 0x05, which is valid for transferOut
            await self.js_dev.transferOut(EP_OUT_ADDR, js_data)

        run_sync(_async_write())

    def read(self, length: int, timeout_ms: int) -> bytes:
        if not self.js_dev:
            raise Exception("Device not open")

        from pyodide.ffi import run_sync

        async def _async_read():
            # EP_IN_ADDR is 0x84. WebUSB expects the Endpoint Number (4).
            # We strip the direction bit (0x80) just in case, though some browsers handle 0x84.
            endpoint_num = EP_IN_ADDR & 0x0F 
            
            result = await self.js_dev.transferIn(endpoint_num, length)
            # result.data is a DataView. buffer.to_bytes() converts it back to Python.
            return result.data.buffer.to_bytes()

        return run_sync(_async_read())

    def close(self):
        if not self.js_dev:
            return

        from pyodide.ffi import run_sync

        async def _async_close():
            try:
                await self.js_dev.close()
            except:
                pass 

        run_sync(_async_close())
        self.js_dev = None